@@ -1,5 +1,5 @@
 # localslackirc
-# Copyright (C) 2018-2021 Salvo "LtWorf" Tomaselli
+# Copyright (C) 2018-2022 Salvo "LtWorf" Tomaselli
 #
 # localslackirc is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -34,30 +34,13 @@ import sys
 import time
 
 
-try:
-    from emoji import emojize  # type: ignore
-except ModuleNotFoundError:
-    def emojize(string:str, use_aliases:bool=False, delimiters: Tuple[str,str]=(':', ':')) -> str:  # type: ignore
-        return string
-
-
 import slack
 from log import *
+import msgparsing
+from diff import seddiff
 
 
-# How slack expresses mentioning users
-_MENTIONS_REGEXP = re.compile(r'<@([0-9A-Za-z]+)>')
-_CHANNEL_MENTIONS_REGEXP = re.compile(r'<#([A-Z0-9]+)\|[_\w-]+>')
-_URL_REGEXP = re.compile(r'<([a-z0-9\-\.]+)://([^\s\|]+)[\|]{0,1}([^<>]*)>')
-
-VERSION = '1.14'
-
-
-_SLACK_SUBSTITUTIONS = [
-    ('&amp;', '&'),
-    ('&gt;', '>'),
-    ('&lt;', '<'),
-]
+VERSION = '1.15'
 
 
 class IrcDisconnectError(Exception): ...
@@ -129,6 +112,7 @@ class Client:
         self.username = b''
         self.realname = b''
         self.parted_channels: Set[bytes] = settings.ignored_channels
+        self.known_threads: Dict[bytes, slack.MessageThread] = {}
         self.hostname = gethostname().encode('utf8')
 
         self.settings = settings
@@ -139,11 +123,6 @@ class Client:
         self._mentions_regex_cache: Dict[str, Optional[re.Pattern]] = {}  # Cache for the regexp to perform mentions. Key is channel id
         self._annoy_users: Dict[str, int] = {} # Users to annoy pretending to type when they type
 
-        if self.settings.provider == Provider.SLACK:
-            self.substitutions = _SLACK_SUBSTITUTIONS
-        else:
-            self.substitutions = []
-
     async def _nickhandler(self, cmd: bytes) -> None:
         if b' ' not in cmd:
             self.nick = b'localslackirc'
@@ -241,10 +220,12 @@ class Client:
         except Exception:
             await self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unable to join channel: {channel_name}')
 
-    async def _send_chan_info(self, channel_name: bytes, slchan: slack.Channel):
+    async def _send_chan_info(self, channel_name: bytes, slchan: Union[slack.Channel, slack.MessageThread]):
         if not self.settings.nouserlist:
+            l = await self.sl_client.get_members(slchan.id)
+
             userlist: List[bytes] = []
-            for i in await self.sl_client.get_members(slchan.id):
+            for i in l:
                 try:
                     u = await self.sl_client.get_user(i)
                 except Exception:
@@ -258,9 +239,10 @@ class Client:
 
             users = b' '.join(userlist)
 
+        topic = (await self.parse_message(slchan.real_topic, b'')).replace('\n', ' | ')
         self.s.write(b':%s!%s@127.0.0.1 JOIN %s\r\n' % (self.nick, self.nick, channel_name))
         await self.s.drain()
-        await self._sendreply(Replies.RPL_TOPIC, slchan.real_topic, [channel_name])
+        await self._sendreply(Replies.RPL_TOPIC, topic, [channel_name])
         await self._sendreply(Replies.RPL_NAMREPLY, b'' if self.settings.nouserlist else users, ['=', channel_name])
         await self._sendreply(Replies.RPL_ENDOFNAMES, 'End of NAMES list', [channel_name])
 
@@ -278,32 +260,40 @@ class Client:
         else:
             action = False
 
-        if dest.startswith(b'#'):
-            to_channel = True
-            dest_object: Union[slack.User, slack.Channel] = await self.sl_client.get_channel_by_name(dest[1:].decode())
+        if dest in self.known_threads:
+            dest_object: Union[slack.User, slack.Channel, slack.MessageThread] = self.known_threads[dest]
+        elif dest.startswith(b'#'):
+            try:
+                dest_object = await self.sl_client.get_channel_by_name(dest[1:].decode())
+            except KeyError:
+                await self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unknown channel {dest.decode()}')
+                return
         else:
-            to_channel = False
-            dest_object = await self.sl_client.get_user_by_name(dest.decode())
-
+            try:
+                dest_object = await self.sl_client.get_user_by_name(dest.decode())
+            except KeyError:
+                await self._sendreply(Replies.ERR_NOSUCHNICK, f'Unknown user {dest.decode()}')
+                return
 
         message = await self._addmagic(msg.decode('utf8'), dest_object)
 
-        if to_channel:
-            await self.sl_client.send_message(
-                dest_object.id,
+        if isinstance(dest_object, slack.User):
+            await self.sl_client.send_message_to_user(
+                dest_object,
                 message,
                 action,
             )
         else:
-            await self.sl_client.send_message_to_user(
-                dest_object.id,
+            await self.sl_client.send_message(
+                dest_object,
                 message,
                 action,
             )
 
     async def _listhandler(self, cmd: bytes) -> None:
         for c in await self.sl_client.channels(refresh=True):
-            await self._sendreply(Replies.RPL_LIST, c.real_topic, ['#' + c.name, str(c.num_members)])
+            topic = (await self.parse_message(c.real_topic, b'')).replace('\n', ' | ')
+            await self._sendreply(Replies.RPL_LIST, topic, ['#' + c.name, str(c.num_members)])
         await self._sendreply(Replies.RPL_LISTEND, 'End of LIST')
 
     async def _modehandler(self, cmd: bytes) -> None:
@@ -337,29 +327,36 @@ class Client:
         #/sendfile #destination filename
         params = cmd.split(b' ', 2)
         try:
+            bchannel_name = params[1]
             channel_name = params[1].decode('utf8')
             filename = params[2].decode('utf8')
         except IndexError:
             await self._sendreply(Replies.ERR_UNKNOWNCOMMAND, 'Syntax: /sendfile #channel filename')
             return
 
-        try:
-            if channel_name.startswith('#'):
-                dest = (await self.sl_client.get_channel_by_name(channel_name[1:])).id
-            else:
-                dest = (await self.sl_client.get_user_by_name(channel_name)).id
-        except KeyError:
-            await self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unable to find destination: {channel_name}')
-            return
+        if bchannel_name in self.known_threads:
+            dest_channel = self.known_threads[bchannel_name]
+            dest = dest_channel.id
+            thread_ts = dest_channel.thread_ts
+        else:
+            thread_ts = None
+            try:
+                if channel_name.startswith('#'):
+                    dest = (await self.sl_client.get_channel_by_name(channel_name[1:])).id
+                else:
+                    dest = (await self.sl_client.get_user_by_name(channel_name)).id
+            except KeyError:
+                await self._sendreply(Replies.ERR_NOSUCHCHANNEL, f'Unable to find destination: {channel_name}')
+                return
 
         try:
-            await self.sl_client.send_file(dest, filename)
+            await self.sl_client.send_file(dest, filename, thread_ts)
             await self._sendreply(0, 'Upload of %s completed' % filename)
         except Exception as e:
             await self._sendreply(Replies.ERR_FILEERROR, f'Unable to send file {e}')
 
     async def _parthandler(self, cmd: bytes) -> None:
-        _, name = cmd.split(b' ', 1)
+        name = cmd.split(b' ')[1]
         self.parted_channels.add(name)
 
     async def _awayhandler(self, cmd: bytes) -> None:
@@ -544,7 +541,7 @@ class Client:
         Adds magic codes and various things to
         outgoing messages
         """
-        for i in self.substitutions:
+        for i in msgparsing.SLACK_SUBSTITUTIONS:
             msg = msg.replace(i[1], i[0])
         if self.settings.provider == Provider.SLACK:
             msg = msg.replace('@here', '<!here>')
@@ -565,94 +562,85 @@ class Client:
                 msg = msg[0:m.start()] + '<@%s>' % (await self.sl_client.get_user_by_name(username)).id + msg[m.end():]
         return msg
 
-    async def parse_message(self, i: str, source: bytes) -> bytes:
-        # Replace all mentions with @user
-        while mention := _MENTIONS_REGEXP.search(i):
-            i = (
-                i[0:mention.span()[0]] +
-                (await self.sl_client.get_user(mention.groups()[0])).name +
-                i[mention.span()[1]:]
-            )
+    async def parse_message(self, i: str, source: bytes) -> str:
+        """
+        This converts a slack message into a message for IRC.
 
-        # Replace all channel mentions
-        while mention := _CHANNEL_MENTIONS_REGEXP.search(i):
-            i = (
-                i[0:mention.span()[0]] +
-                '#' +
-                (await self.sl_client.get_channel(mention.groups()[0])).name_normalized +
-                i[mention.span()[1]:]
-            )
+        It will replace mentions and shouts with the IRC equivalent.
 
-        #Replace URL with bottom links
-        bottom = ""
+        It will save preformatted text into txt files and link them
+        if the settings are such.
+
+        It will put the links at the end like with emails.
+        """
+
+        r = ''
+
+        # Url replacing
+        links = ''
         refs = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
         refn = 1
-        while url := _URL_REGEXP.search(i):
-            schema, path, label = url.groups()
-            if label and (label != f"{schema}://{path}"):
-                if '://' in label:
-                    label = 'LINK'
-                ref = str(refn).translate(refs)
-                refn += 1
-                i = (
-                    i[0:url.span()[0]] +
-                    f'{label}{ref}' +
-                    i[url.span()[1]:]
-                )
-                bottom += f'\n  {ref} {schema}://{path}'
-            else:
-                i = (
-                    i[0:url.span()[0]] +
-                    f'{schema}://{path}' +
-                    i[url.span()[1]:]
-                )
-        i += bottom
 
-        for s in self.substitutions:
-            i = i.replace(s[0], s[1])
-
-        # Replace emoji codes (e.g. :thumbsup:)
-        i = emojize(i, use_aliases=True)
-
-        # Store long formatted text into txt files
-        if self.settings.formatted_max_lines:
-            begin = 0
-            while True:
-                try:
-                    fmtstart = i.index('```', begin)
-                    fmtend = i.index('```', fmtstart + 1)
-
-                    formatted = i[fmtstart + 3:fmtend]
-                    prefix = i[0:fmtstart]
-                    suffix = i[fmtend + 3:]
-
-                    if formatted.count('\n') > self.settings.formatted_max_lines:
-                        import tempfile
-                        with tempfile.NamedTemporaryFile(
-                                    mode='wt',
-                                    dir=self.settings.downloads_directory,
-                                    suffix='.txt',
-                                    prefix='localslackirc-attachment-',
-                                    delete=False) as tmpfile:
-                            tmpfile.write(formatted)
-                            i = prefix + f'\n === PREFORMATTED TEXT AT file://{tmpfile.name}\n' + suffix
+        for t in msgparsing.tokenize(i):
+            if isinstance(t, str): # A normal nice string
+                r += t
+            elif isinstance(t, msgparsing.PreBlock): # Preformatted block
+                # Store long formatted text into txt files
+                if self.settings.formatted_max_lines and t.lines > self.settings.formatted_max_lines:
+                    import tempfile
+                    with tempfile.NamedTemporaryFile(
+                            mode='wt',
+                            dir=self.settings.downloads_directory,
+                            suffix='.txt',
+                            prefix='localslackirc-attachment-',
+                            delete=False) as tmpfile:
+                        tmpfile.write(t.txt)
+                        r += f'\n === PREFORMATTED TEXT AT file://{tmpfile.name}\n'
+                else: # Do not store to file
+                    r += f'```{t.txt}```'
+            elif isinstance(t, msgparsing.SpecialItem):
+                if t.kind == msgparsing.Itemkind.MENTION: # User mention
+                    r += (await self.sl_client.get_user(t.val)).name
+                elif t.kind == msgparsing.Itemkind.CHANNEL: # Channel mention
+                    r += '#' + (await self.sl_client.get_channel(t.val)).name_normalized
+                elif t.kind == msgparsing.Itemkind.YELL: # Channel shouting
+                    yell = ' [%s]:' % self.nick.decode('utf8') if source not in self.settings.silenced_yellers else ':'
+                    if t.val == 'here':
+                        r += 'yelling' + yell
+                    elif t.val == 'channel':
+                        r += 'YELLING LOUDER' + yell
                     else:
-                        i = '```'.join((prefix, formatted, suffix))
-                        begin = fmtend + 1
-                except ValueError:
-                    break
+                        r += 'DEAFENING YELL' + yell
+                else: # Link
+                    label = t.human
+                    if label is None:
+                        r += t.val
+                    else:
+                        if '://' in label:
+                            label = 'LINK'
+                        ref = str(refn).translate(refs)
+                        links += f'\n  {ref} {t.val}'
+                        r += label + ref
+                        refn += 1
+        return r + links
 
-        encoded = i.encode('utf8')
+    async def _messageedit(self, sl_ev: slack.MessageEdit) -> None:
+        if not sl_ev.is_changed:
+            return
+        source = (await self.sl_client.get_user(sl_ev.previous.user)).name.encode('utf8')
+        previous = await self.parse_message(sl_ev.previous.text, source)
+        current = await self.parse_message(sl_ev.current.text, source)
 
-        if self.settings.provider == Provider.SLACK:
-            yell = b' [%s]:' % self.nick if source not in self.settings.silenced_yellers else b':'
-            encoded = encoded.replace(b'<!here>', b'yelling%s' % yell)
-            encoded = encoded.replace(b'<!channel>', b'YELLING LOUDER%s' % yell)
-            encoded = encoded.replace(b'<!everyone>', b'DEAFENING YELL%s' % yell)
+        diffmsg = slack.Message(
+            text=seddiff(sl_ev.previous.text, sl_ev.current.text),
+            channel=sl_ev.channel,
+            user=sl_ev.previous.user,
+            thread_ts=sl_ev.previous.thread_ts
+        )
 
-        return encoded
+        await self._message(diffmsg)
 
-    async def _message(self, sl_ev: Union[slack.Message, slack.MessageDelete, slack.MessageBot, slack.ActionMessage], prefix: str=''):
+    async def _message(self, sl_ev: Union[slack.Message, slack.MessageDelete, slack.MessageBot, slack.ActionMessage], prefix: str='') -> None:
         """
         Sends a message to the irc client
         """
@@ -660,6 +648,7 @@ class Client:
             source = (await self.sl_client.get_user(sl_ev.user)).name.encode('utf8')
         else:
             source = b'bot'
+
         try:
             dest = b'#' + (await self.sl_client.get_channel(sl_ev.channel)).name.encode('utf8')
         except KeyError:
@@ -671,9 +660,25 @@ class Client:
             # Ignoring messages, channel was left on IRC
             return
 
+        if sl_ev.thread_ts:
+            # Threaded message
+            thread = await self.sl_client.get_thread(sl_ev.thread_ts, sl_ev.channel)
+            dest = b'#' + thread.name.encode('utf8')
+
+            # Join thread channel if needed
+            if dest not in self.known_threads or dest in self.parted_channels:
+                if dest in self.parted_channels:
+                    self.parted_channels.remove(dest)
+                await self._send_chan_info(dest, self.known_threads.get(dest, thread))
+                self.known_threads[dest] = self.known_threads.get(dest, thread)
+
         text = sl_ev.text
 
-        lines = await self.parse_message(prefix + text, source)
+        if sl_ev.files:
+            for f in sl_ev.files:
+                text+=f'\n[file upload] {f.name}\n{f.mimetype} {f.size} bytes\n{f.url_private}'
+
+        lines = (await self.parse_message(prefix + text, source)).encode('utf-8')
         for i in lines.split(b'\n'):
             if not i:
                 continue
@@ -722,13 +727,9 @@ class Client:
         elif isinstance(sl_ev, slack.ActionMessage):
             await self._message(sl_ev)
         elif isinstance(sl_ev, slack.MessageEdit):
-            if sl_ev.is_changed:
-                await self._message(sl_ev.diffmsg)
+            await self._messageedit(sl_ev)
         elif isinstance(sl_ev, slack.MessageBot):
             await self._message(sl_ev, '[%s] ' % sl_ev.username)
-        elif isinstance(sl_ev, slack.FileShared):
-            f = await self.sl_client.get_file(sl_ev)
-            await self._message(f.announce())
         elif isinstance(sl_ev, slack.Join):
             await self._joined_parted(sl_ev, True)
         elif isinstance(sl_ev, slack.Leave):
