@@ -1,5 +1,5 @@
 # localslackirc
-# Copyright (C) 2018-2021 Salvo "LtWorf" Tomaselli
+# Copyright (C) 2018-2022 Salvo "LtWorf" Tomaselli
 #
 # localslackirc is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,7 +25,6 @@ from typing import *
 
 from typedload import load, dump
 
-from diff import seddiff
 from slackclient import SlackClient
 from slackclient.client import LoginInfo
 from log import *
@@ -44,7 +43,9 @@ USELESS_EVENTS = {
     'file_deleted',
     'file_public',
     'file_created',
+    'file_shared',
     'desktop_notification',
+    'mobile_in_app_notification',
 }
 
 
@@ -62,6 +63,17 @@ class Response(NamedTuple):
     error: Optional[str] = None
 
 
+@dataclass
+class File:
+    id: str
+    url_private: str
+    size: int
+    user: str
+    name: Optional[str] = None
+    title: Optional[str] = None
+    mimetype: Optional[str] = None
+
+
 class Topic(NamedTuple):
     """
     In slack, topic is not just a string, but has other fields.
@@ -77,7 +89,8 @@ class LatestMessage(NamedTuple):
         return datetime.datetime.utcfromtimestamp(self.ts)
 
 
-class Channel(NamedTuple):
+@dataclass(frozen=True)
+class Channel:
     """
     A channel description.
 
@@ -108,18 +121,26 @@ class Channel(NamedTuple):
             t = self.topic.value
         else:
             t = self.purpose.value
-        return t.replace('\n', ' | ')
+        return t
+
+@dataclass(frozen=True)
+class MessageThread(Channel):
+    thread_ts: str = ''
 
 
-class Message(NamedTuple):
+@dataclass(frozen=True)
+class Message:
     channel: str  # The channel id
     user: str  # The user id
     text: str
+    thread_ts: Optional[str] = None
+    files: List[File] = field(default_factory=list)
 
 
 class NoChanMessage(NamedTuple):
     user: str
     text: str
+    thread_ts: Optional[str] = None
 
 
 class ActionMessage(Message):
@@ -144,14 +165,6 @@ class MessageEdit:
     def is_changed(self) -> bool:
         return self.previous.text != self.current.text
 
-    @property
-    def diffmsg(self) -> Message:
-        return Message(
-            text=seddiff(self.previous.text, self.current.text),
-            channel=self.channel,
-            user=self.current.user,
-        )
-
 
 @dataclass
 class MessageDelete:
@@ -159,6 +172,11 @@ class MessageDelete:
     subtype: Literal['message_deleted']
     channel: str
     previous_message: NoChanMessage
+    files: List[File] = field(default_factory=list)
+
+    @property
+    def thread_ts(self) -> Optional[str]:
+        return self.previous_message.thread_ts
 
     @property
     def user(self) -> str:
@@ -183,43 +201,6 @@ class Profile(NamedTuple):
     is_ultra_restricted: bool = False
 
 
-@dataclass
-class File:
-    id: str
-    url_private: str
-    size: int
-    user: str
-    name: Optional[str] = None
-    title: Optional[str] = None
-    mimetype: Optional[str] = None
-    channels: List[str] = field(default_factory=list)
-    groups: List[str] = field(default_factory=list)
-    ims: List[str] = field(default_factory=list)
-
-    def announce(self) -> Message:
-        """
-        Returns a message to announce this file.
-        """
-        return Message(
-            channel=(self.channels + self.groups + self.ims).pop(),
-            user=self.user,
-            text='[file upload] %s\n%s %d bytes\n%s' % (
-                self.name,
-                self.mimetype,
-                self.size,
-                self.url_private
-            )
-        )
-
-
-@dataclass
-class FileShared:
-    type: Literal['file_shared']
-    file_id: str
-    user_id: str
-    ts: float
-
-
 @dataclass
 class MessageBot:
     type: Literal['message']
@@ -229,6 +210,8 @@ class MessageBot:
     channel: str
     bot_id: Optional[str] = None
     attachments: List[Dict[str, Any]] = field(default_factory=list)
+    thread_ts: Optional[str] = None
+    files: List[File] = field(default_factory=list)
 
     @property
     def text(self):
@@ -328,7 +311,6 @@ SlackEvent = Union[
     Message,
     ActionMessage,
     MessageBot,
-    FileShared,
     Join,
     Leave,
     GroupJoined,
@@ -382,6 +364,17 @@ class Slack:
         log('Login in slack')
         self.login_info = await self.client.login(15)
 
+    async def get_history(self, channel: Union[Channel, IM, str], ts: str, cursor: Optional[NextCursor]=None, limit: int=1000, inclusive: bool=False) -> History:
+        p = await self.client.api_call(
+            'conversations.history',
+            channel=channel if isinstance(channel, str) else channel.id,
+            oldest=ts,
+            limit=limit,
+            cursor=cursor.next_cursor if cursor else None,
+            inclusive=inclusive,
+        )
+        return load(p, History)
+
     async def _thread_history(self, channel: str, thread_id: str) -> List[Union[HistoryMessage, HistoryBotMessage]]:
         r: List[Union[HistoryMessage, HistoryBotMessage]] = []
         cursor = None
@@ -443,18 +436,10 @@ class Slack:
             cursor = None
             while True: # Loop to iterate the cursor
                 log('Calling cursor')
-                r = await self.client.api_call(
-                    'conversations.history',
-                    channel=channel.id,
-                    oldest=last_timestamp,
-                    limit=1000,
-                    cursor=cursor,
-                )
                 try:
-                    response = load(r, History)
+                    response = await self.get_history(channel, str(last_timestamp))
                 except Exception as e:
                     log('Failed to parse', e)
-                    log(r)
                     break
                 msg_list = list(response.messages)
                 while msg_list:
@@ -467,13 +452,8 @@ class Slack:
                     if self._status.last_timestamp < msg.ts:
                         self._status.last_timestamp = msg.ts
 
-                    # The attached files
-                    for f in msg.files:
-                        f.channels.append(channel.id)
-                        self._internalevents.append(f.announce())
-
                     # History for the thread
-                    if  msg.thread_ts and float(msg.thread_ts) == msg.ts:
+                    if msg.thread_ts and float(msg.thread_ts) == msg.ts:
                         l = await self._thread_history(channel.id, msg.thread_ts)
                         l.reverse()
                         msg_list = l + msg_list
@@ -484,7 +464,9 @@ class Slack:
                         self._internalevents.append(Message(
                             channel=channel.id,
                             text=msg.text,
-                            user=msg.user
+                            user=msg.user,
+                            thread_ts=msg.thread_ts,
+                            files=msg.files,
                         ))
                     elif isinstance(msg, HistoryBotMessage):
                         self._internalevents.append(MessageBot(
@@ -494,6 +476,7 @@ class Slack:
                             username=msg.username,
                             channel=channel.id,
                             bot_id=msg.bot_id,
+                            thread_ts=msg.thread_ts,
                         ))
 
                 if response.has_more and response.response_metadata:
@@ -656,6 +639,34 @@ class Slack:
                     return c
         raise KeyError()
 
+    async def get_thread(self, thread_ts: str, original_channel: str) -> MessageThread:
+        """
+        Creates a fake channel class for a chat thread
+        """
+        channel = (await self.get_channel(original_channel)).name_normalized
+
+        # Get head message
+        history = await self.get_history(original_channel, thread_ts, None, 1, True)
+
+        msg = history.messages.pop()
+        user = (await self.get_user(msg.user)).name if isinstance(msg, HistoryMessage) else 'bot'
+
+        # Top message is a file
+        if msg.text == '' and msg.files:
+            f = msg.files[0]
+            original_txt = f'{f.title} {f.mimetype} {f.url_private}'
+        else:
+            original_txt = msg.text.strip().replace('\n', ' | ')
+
+        t = Topic(f'{user} in {channel}: {original_txt}')
+        return MessageThread(
+            id=original_channel,
+            name_normalized=f'threaded-{thread_ts}',
+            purpose=t,
+            topic=t,
+            thread_ts=thread_ts,
+        )
+
     async def get_im(self, im_id: str) -> Optional[IM]:
         if not im_id.startswith('D'):
             return None
@@ -723,19 +734,7 @@ class Slack:
         else:
             raise KeyError(response)
 
-    async def get_file(self, f: Union[FileShared, str]) -> File:
-        """
-        Returns a file object
-        """
-        fileid = f if isinstance(f, str) else f.file_id
-        r = await self.client.api_call("files.info", file=fileid)
-        response = load(r, Response)
-        if response.ok:
-            return load(r['file'], File)
-        else:
-            raise KeyError(response)
-
-    async def send_file(self, channel_id: str, filename: str) -> None:
+    async def send_file(self, channel_id: str, filename: str, thread_ts: Optional[str]) -> None:
         """
         Send a file to a channel or group or whatever
         """
@@ -743,6 +742,7 @@ class Slack:
             r = await self.client.api_call(
                 'files.upload',
                 channels=channel_id,
+                thread_ts=thread_ts,
                 file=f,
             )
         response = load(r, Response)
@@ -762,7 +762,11 @@ class Slack:
         for i in r:
             self._sent_by_self.remove(i)
 
-    async def send_message(self, channel_id: str, msg: str, action: bool) -> None:
+    async def send_message(self, channel: Union[Channel, MessageThread], msg: str, action: bool) -> None:
+        thread_ts = channel.thread_ts if isinstance(channel, MessageThread) else None
+        return await self._send_message(channel.id, msg, action, thread_ts)
+
+    async def _send_message(self, channel_id: str, msg: str, action: bool, thread_ts: Optional[str]) -> None:
         """
         Send a message to a channel or group or whatever
         """
@@ -772,12 +776,18 @@ class Slack:
             api = 'chat.postMessage'
 
         try:
+            kwargs = {}
+
+            if thread_ts:
+                kwargs['thread_ts'] = thread_ts
+
             self._wsblock += 1
             r = await self.client.api_call(
                 api,
                 channel=channel_id,
                 text=msg,
                 as_user=True,
+                **kwargs,  # type: ignore
             )
             response = load(r, Response)
             if response.ok and response.ts:
@@ -787,7 +797,7 @@ class Slack:
         finally:
             self._wsblock -= 1
 
-    async def send_message_to_user(self, user_id: str, msg: str, action: bool):
+    async def send_message_to_user(self, user: User, msg: str, action: bool):
         """
         Send a message to a user, pass the user id
         """
@@ -796,15 +806,15 @@ class Slack:
         # so to deliver a message to them, a channel id is required.
         # Those are called IM.
 
-        if user_id in self._imcache:
+        if user.id in self._imcache:
             # channel id is cached
-            channel_id = self._imcache[user_id]
+            channel_id = self._imcache[user.id]
         else:
             # Find the channel id
             found = False
             # Iterate over all the existing conversations
             for i in await self.get_ims():
-                if i.user == user_id:
+                if i.user == user.id:
                     channel_id = i.id
                     found = True
                     break
@@ -813,16 +823,16 @@ class Slack:
                 r = await self.client.api_call(
                     "im.open",
                     return_im=True,
-                    user=user_id,
+                    user=user.id,
                 )
                 response = load(r, Response)
                 if not response.ok:
                     raise ResponseException(response.error)
                 channel_id = r['channel']['id']
 
-            self._imcache[user_id] = channel_id
+            self._imcache[user.id] = channel_id
 
-        await self.send_message(channel_id, msg, action)
+        await self._send_message(channel_id, msg, action, None)
 
     async def event(self) -> Optional[SlackEvent]:
         """
@@ -859,7 +869,7 @@ class Slack:
                 continue
 
             debug(event)
-            loadable_events = Union[TopicChange, FileShared, MessageBot, MessageEdit, MessageDelete, GroupJoined, Join, Leave, UserTyping]
+            loadable_events = Union[TopicChange, MessageBot, MessageEdit, MessageDelete, GroupJoined, Join, Leave, UserTyping]
             try:
                 ev: Optional[loadable_events] = load(
                     event,
@@ -888,9 +898,9 @@ class Slack:
                     # the other user, and prepend them with "I say: "
                     im = await self.get_im(msg.channel)
                     if im and im.user != msg.user:
-                        msg = Message(user=im.user, text='I say: ' + msg.text, channel=im.id)
+                        msg = Message(user=im.user, text='I say: ' + msg.text, channel=im.id, thread_ts=msg.thread_ts)
                     if subt == 'me_message':
-                        return ActionMessage(*msg)
+                        return ActionMessage(*msg)  # type: ignore
                     else:
                         return msg
                 elif t == 'message' and subt == 'slackbot_response':
